; Useful, common functions.       -*- mode: scheme; mode: font-lock -*-
; 2000-07-24 / lth

(require 'list)
(require 'generator)

(define (symbol<? a b)
  (string<? (symbol->string a) (symbol->string b)))

(define (benchmark<? a b)
  (or (symbol<? (cadr a) (cadr b))
      (and (eq? (cadr a) (cadr b))
           (< (car a) (car b)))))


; (sum numbers) == (apply + numbers)
; (sum fn l1 l2 ...) == (apply + (map fn l1 l2 ...))

(define (sum a1 . args)
  (if (null? args) 
      (apply + a1)
      (apply + (apply map a1 args))))


(define (square x) (* x x))


(define (mean numbers)
  (/ (sum numbers) (length numbers)))


(define (weighted-mean data get-val get-weight)
  (/ (sum (lambda (x) (* (get-val x) (get-weight x))) data) 
     (sum get-weight data)))


(define (median numbers)
  (let ((n (length numbers)))
    (if (odd? n)
        (list-ref numbers (quotient n 2))
        (/ (+ (list-ref numbers (quotient n 2))
              (list-ref numbers (- (quotient n 2) 1)))
           2))))


(define (standard-deviation numbers)
  (let ((avg (mean numbers)))
    (sqrt (/ (sum (lambda (x) (square (- x avg))) numbers)
             (length numbers)))))


; Given a list of data, and functions to extract the X and Y values 
; from each datum, return two values a and b such that Y=a+bX is the
; least-squares regression line through the data points.  (That is,
; it returns y-intercept and slope _in that order_.)

(define (least-squares data get-x get-y)
  (let ((n  (length data))
        (xs (map get-x data))
        (ys (map get-y data)))
    (let* ((b (/ (- (* n (sum * xs ys)) (* (sum xs) (sum ys)))
                 (- (* n (sum square xs)) (square (sum xs)))))
           (a (/ (- (sum ys) (* b (sum xs)))
                 n)))
      (values a b))))


; Return the data in the interquartile range from a data set 
; sorted in some order (think nondecreasing).

(define (interquartile datum-value data)
  (let ((total (sum datum-value data)))
    (let loop ((sum 0) (data data))
      (if (< (+ sum (datum-value (car data))) (* 0.25 total))
          (loop (+ sum (datum-value (car data))) (cdr data))
          (let ((start data))
            (let loop ((sum sum) (data data) (n 0))
              (if (> sum (* 0.75 total))
                  (list-head start n)
                  (loop (+ sum (datum-value (car data)))
                        (cdr data) 
                        (+ n 1)))))))))


; The _datum_ is as produced by basic-statistics.

(define (benchmark-designator datum)
  (benchmark-designator* (lambda (tag)
                           (let ((probe (assq tag datum)))
                             (if probe
                                 (cdr probe)
                                 (error "benchmark-designator: can't find " 
                                        tag))))))

; Like above, but get is a getter function

(define (benchmark-designator* get)
  (list (cadr (assq 'iterations_per_run (get 'system-settings)))
        (get 'benchmark-name)))


(define (format-benchmark-name name)

  (define (rationalize-name x)
    (cond ((eq? x 'gcold:25_0_1_0_200) "gcold:1,0")
          ((eq? x 'gcold:25_0_1_1000_200) "gcold:1,1000")
          ((eq? x 'gcold:25_0_10_1000_200) "gcold:10,1000")
          ((eq? x 'perm:200_8_10_1) "perm:200,1")
          ((eq? x 'perm:25_8_10_8) "perm:25,8")
          (else x)))

  (format #f "{\\tt ~a~a}" 
          (if (= 1 (car name)) 
              "" 
              (car name)) 
          (rationalize-name (cadr name))))


; The _datum_ is as produced by basic-statistics.

(define (gc-designator datum)
  (let ((x (assq 'system-info datum)))
    (if x
        (cadr x)
        #f)))

; The _datum_ is as produced by basic-statistics.

(define (all-iterations-succeeded? item)
  (and (assq 'bytes-max-heap item)
       (not (cdr (assq 'some-failed item)))))


; Basic notion of comparable runs: Two runs are comparable if they are
; for the same benchmark and either are both fixed-heap with the same
; upper and lower limits and trigger, or variable-heap with the same
; load factor.
;
; _d1_ and _d2_ are single data as produced by basic-statistics.

(define (comparable-runs? d1 d2)
  (and (equal? (benchmark-designator d1)
               (benchmark-designator d2))
       (let ((s1 (assq 'system-info d1))
             (s2 (assq 'system-info d2)))
         (and (eq? (caddr s1) (caddr s2)) ; upper, upper+lower, variable
              (= (cadddr s1) (cadddr s2)) ; load or heap upper limit
              (or (eq? (caddr s1) 'variable)
                  (= (car (cddddr s1)) (car (cddddr s2)))) ; trigger
              ))))


; Select the variable-heap runs with the given load factor.
; Data are lists of data, as returned by basic-statistics.

(define (select-load-factor load data . rest)
  (let ((pred (if (null? rest) = (car rest))))
    (filter (lambda (datum)
              (let ((x (assq 'system-info datum)))
                (and (eq? 'variable (caddr x))
                     (pred (cadddr x) load))))
            data)))

(define (select-trigger load data . rest)
  (let ((pred (if (null? rest) = (car rest))))
    (filter (lambda (datum)
              (let ((x (assq 'system-info datum)))
                (and (memq (caddr x) '(upper upper+lower fixed))
                     (pred (car (cddddr x)) load))))
            data)))

(define (load/trigger=? datum n)
  (= n (get-load/trigger datum)))

(define (get-load/trigger datum)
  (let ((info (assq 'system-info datum)))
    (cond ((eq? 'variable (caddr info))
           (cadddr info))
          ((memq (caddr info) '(upper upper+lower fixed))
           (car (cddddr info)))
          (else
           (error "get-load/trigger: unknown run type: " info)))))

(define (get-trigger datum)
  (let ((info (assq 'system-info datum)))
    (cond ((eq? 'variable (caddr info))
           0)
          ((memq (caddr info) '(upper upper+lower fixed))
           (car (cddddr info)))
          (else
           (error "get-trigger: unknown run type: " info)))))

; Process a list as returned by *-canonical-data, and apply p to
; each datum and to the n that goes with it.
;
; data ::= ( ( basic-statistics n ) ... )

(define (for-each-benchmark-result process-datum data)
  (do ((data data (cdr data)))
      ((null? data))
    (let* ((x (car data))
           (n (cadr x))
           (bs (car x)))
      (do ((bs bs (cdr bs)))
          ((null? bs))
        (process-datum (car bs) n)))))

; This is like the one above, but it takes several lists of
; data, and traverses them all in parallel and applies the
; processor to n and all the data (NB note change of order).
; The lists had better have exactly the same structure!!
; It's in principle compatible with the one above, but since
; the arguments are most naturally passed in a different order,
; I gave it a new name.

(define (for-each-benchmark-result* process-datum . all-data)
  (let* ((end (list 'END))
         (generators
          (map (lambda (data)
                 (make-generator 
                  (lambda (return)
                    (for-each-benchmark-result (lambda (datum n)
                                                 (return (cons datum n)))
                                               data)
                    (return END))))
               all-data)))
    (let loop ()
      (let ((xs (map (lambda (x) (x)) generators)))
        (if (not (eq? (car xs) END))
            (begin (apply process-datum (cdar xs) (map car xs))
                   (loop))
            (if (not (every? (lambda (x) (eq? x END)) xs))
                (error "for-each-benchmark-result*: dissimilar structures.")))))))


; As for-each-benchmark-result, but return list of results returned 
; by function.  Do not accumulate #f results.

(define (filter-map-benchmark-result process-datum data)
  (call-with-accumulator
   (lambda (acc)
     (for-each-benchmark-result
      (lambda (d n)
        (let ((x (process-datum d n)))
          (if x (acc x))))
      data))))

; Given a benchmark designator and a crunched but otherwise unprocessed
; datum, remove all mention of the benchmark from the datum and return
; the datum.

(define (remove-benchmark bm datum)
  (filter (lambda (x)
            (or (= 1 (length (system.benchmark-name x)))
                (error "Invariant violated in remove-benchmark."))
            (let ((thisbm (list (cadr (assq 'iterations_per_run 
                                            (system.settings x)))
                                (car (system.benchmark-name x)))))
              (not (equal? thisbm bm))))
          datum))

(define (remove-benchmarks bms datum)
  (if (null? bms)
      datum
      (remove-benchmarks (cdr bms)
                         (remove-benchmark (car bms) datum))))

; Remove a system with the option from a crunched but otherwised
; unprocessed datum.  Option can be an atom or a list of atoms.

(define (remove-system option datum)
  (if (atom? option)
      (filter (lambda (x)
                (not (memv option (system.info x))))
              datum)
      (filter (lambda (x)
                (not (sublistv option (system.info x))))
              datum)))

; Just a handy table that maps benchmark designators to data.

(define (make-benchmark-associator init)

  (define results '())

  (define (lookup bm)
    (let ((probe (assoc bm results)))
      (if probe
          probe
          (let ((x (cons bm (init))))
            (set! results (cons x results))
            x))))

  (lambda (msg)
    (case msg
      ((lookup) lookup)
      ((results) (lambda () results))
      (else ???))))

(define (sort-by-euclidian-distance-from-origin data get-x get-y)
  (sort data
        (lambda (a b)
          (< (sqrt (+ (square (get-x a)) (square (get-y a))))
             (sqrt (+ (square (get-x b)) (square (get-y b))))))))

(define (sort-by-x-and-y data get-x get-y)
  (sort data
        (lambda (a b)
          (or (< (get-x a) (get-x b))
              (and (= (get-x a) (get-x b))
                   (< (get-y a) (get-y b)))))))

(define (sort-by-benchmark-name data get-name)
  (sort data (lambda (a b) (benchmark<? (get-name a) (get-name b)))))

(define (sort-by-load/trigger data)
  (sort data
        (lambda (a b)
          (< (get-load/trigger a)
             (get-load/trigger b)))))

(define (sort-by-trigger data)          ; And leave VAR data in same order
  (sort data
        (lambda (a b)
          (< (get-trigger a)
             (get-trigger b)))))
                     
(define (sort-by-standard-header data)
  (sort data (lambda (a b)
               (list<? (standard-header a) (standard-header b)))))

; Makes string safe to use in a file name for Mac, Unix, and Windows.

(define (sanitize-string s)
  (list->string
   (map (lambda (x)
          (if (memv x '(#\: #\/ #\\ #\.))
              #\_
              x))
        (string->list s))))

(define (sanitize-string-for-latex s)
  (list->string
   (map (lambda (x)
          (if (memv x '(#\_ #\$))
              #\:
              x))
        (string->list s))))

; DATUM is an assoc list.

(define (standard-header datum)

  (define fixed-runs '(fixed upper upper+lower)) ; tags for fixed-heap runs
  (define variable-runs '(variable))             ; tags for variable-heap runs

  (let* ((sys       (cdr (assq 'system-info datum)))
         (bm        (cdr (assq 'benchmark-name datum)))
         (collector (car sys))
         (heap      (cadr sys))
         (size      (if (memq heap fixed-runs) (caddr sys) 0))
         (trigger   (if (memq heap fixed-runs) (cadddr sys) 0))
         (load      (if (memq heap variable-runs) (caddr sys) 0))
         (aux       (cdddr sys)))
    (list bm collector heap trigger load size aux)))


; DATA is as produced by basic-statistics.
;
; SYSTEM is a system designator list: a list of at least 3, where the
; three first elements should be collector type, run type, and either
; heap size or load factor.
;
; BM is a benchmark name (not a benchmark designator).
;
; Returns the first run in DATA that has benchmark name and system info
; as given.  (system in the list-of-systems corresponding to the systems
; description 'system' and with benchmark name 'bm'.  
;
; (This function is probably obsolete; it is also not very useful.)

(define (find-system data system bm)

  (define (same? a b)
    (and (>= (length a) 3)
         (>= (length b) 3)
         (equal? (car a) (car b))
         (equal? (cadr a) (cadr b))
         (equal? (caddr a) (caddr b))))
    
  (if (null? data)
      (error "No system for " system)
      (let ((s (car data)))
        (if (and (equal? bm (cdr (assq 'benchmark-name s)))
                 (same? system (cdr (assq 'system-info s))))
            s
            (find-system (cdr data) system bm)))))


(define (call-with-output-medium file-or-port proc)
  (cond ((output-port? file-or-port)
         (proc file-or-port))
        ((string? file-or-port)
         (call-with-output-file file-or-port proc))
        (else ???)))


(define (call-with-output-string proc)
  (let ((s (open-output-string)))
    (proc s)
    (get-output-string s)))


(define (decimals n)
  (let ((k (expt 10. n)))
    (lambda (x)
      (/ (round (* x k)) k))))


; Given a list of points (x . y) ordered by nondecreasing x coordinate,
; return two parameters k, and c such that the curve f(L)=(1/(L-k))+c is a 
; good fit for the points, and k < all x.
;
; (This is a feeble attempt at nonlinear regression, since I don't have 
; a text handy to explain how to do that.)

(define (fit-to-mark/cons-curve points)
  (let ((best-fit +inf.0)
        (best-points '(+inf.0 +inf.0))
        (minx (least < (map car points))))
    (do ((k 20 (+ k 1)))
        ((or (= k 100) (>= (/ k 20.0) minx)))
      (do ((c -40 (+ c 1)))
          ((= c 2000))
        (let ((fit (evaluate-fit (/ k 20.0) (/ c 10.0) points)))
          (if (< fit best-fit)
              (begin (set! best-fit fit)
                     (set! best-points (list (/ k 20.0) (/ c 10.0))))))))
    (apply values best-points)))

(define (evaluate-fit k c points)
  (do ((points points (cdr points))
       (ssq 0 (+ ssq (let ((x (caar points))
                           (y (cdar points)))
                       (let ((v (- y (+ (/ 1.0 (- x k)) c))))
                         (* v v))))))
      ((null? points)
       ssq)))

(define (plot-mark/cons-curve out k c lmin lmax)
  (let ((steps 30.0)
        (d (decimals 3)))
    (do ((L lmin (+ L (/ (- lmax lmin) steps))))
        ((>= L lmax))
      (format out "~a ~a~%" (d L) (d (+ (/ 1.0 (- L k)) c))))))


; Data are as returned by basic-statistics.
;
; Returns a list (bm (prom-copy . prom-move) (coll-copy . coll-move))
; for every successful benchmark.

(define (extract-promotion/collection-volume data)
  (filter-map (lambda (x)
                (if (all-iterations-succeeded? x)
                    (list (benchmark-designator x)
                          (cons 
                           (cdr (assq 'run-bytes-copied-promotion-best x))
                           (cdr (assq 'run-bytes-moved-promotion-best x)))
                          (cons
                           (cdr (assq 'run-bytes-copied-collection-best x))
                           (cdr (assq 'run-bytes-moved-collection-best x))))
                    #f))
              data))


; Following procedures from Auxlib/io.sch and Auxlib/string.sch.

(define (read-line . rest)

  (define (finish l k)
    (let ((s (make-string k)))
      (do ((i (- k 1) (- i 1))
	   (l l (cdr l)))
	  ((< i 0) s)
	(string-set! s i (car l)))))

  (define (loop p l k)
    (let ((c (read-char p)))
      (cond ((eof-object? c)
	     (if (null? l)
		 c
		 (finish l k)))
	    ((char=? c #\newline)
	     (finish l k))
	    (else
	     (loop p (cons c l) (+ k 1))))))

  (if (null? rest)
      (loop (current-input-port) '() 0)
      (loop (car rest) '() 0)))

(define (substring-match s subs . rest)
  (let* ((start (if (null? rest) 0 (car rest)))
	 (ls    (string-length s))
	 (lu    (string-length subs))
	 (limit (- ls lu)))
    (let loop ((i start))
      (cond ((> i limit)
	     #f)
	    ((string=? subs (substring s i (+ i lu)))
	     i)
	    (else (loop (+ i 1)))))))


; eof
