; More output processing code!      -*- mode: scheme; mode: font-lock -*-
; 2000-06-30 / lth

(require 'list)

; Data are as returned by basic-statistics.
;
; Removes noise options from the system-info lists (makes printing
; the data less painful) and re-sort the data according to benchmark
; name and option set.  Returns a new list.

(define (remove-option-noise data)

  (define keywords-to-remove
    `((-ssb . #t)
      (-rhash . #t)
      (-dof-growth-divisor . #t)
      (-areas . #t)))

  (define (remove-from sys)
    (if (null? sys)
        '()
        (let ((probe (assq (car sys) keywords-to-remove)))
          (cond (probe
                 (if (cdr probe)
                     (remove-from (cddr sys))
                     (remove-from (cdr sys))))
                (else
                 (cons (car sys)
                       (remove-from (cdr sys))))))))

  (define (less? a b)
    (list<? (cons (cdr (assq 'benchmark-name a))
                  (cdr (assq 'system-info a)))
            (cons (cdr (assq 'benchmark-name b))
                  (cdr (assq 'system-info b)))))

  (for-each (lambda (system)
              (let ((system-info (assq 'system-info system)))
                (set-cdr! system-info (remove-from (cdr system-info)))))
            data)
  (sort data less?))


; Data are as returned by basic-statistics.
;
; Flattens many options from the system-info lists into the datum
; proper, thus adding fields to the assoc lists.  Removes the options
; that are flattened from the system-info field.  Re-sorts the data
; according to benchmark name and system info, and returns a new list.
;
; A keyword that is present but takes no option is added to the
; data as (keyword . #t).  A keyword that is absent and takes no
; options is added to the data as (keyword . #f).  A keyword that
; takes an opion but is not present is not added to the data.

(define (flatten-option-noise data)

  ; A keyword here has #t if it takes an option, #f if it does not.

  (define keywords-to-remove
    `((-ssb . #t)
      (-rhash . #t)
      (-dof-growth-divisor . #t)
      (-dof-fullgc-frequency . #t)
      (-dof-no-shadow-remsets . #f)
      (-dof-fullgc-on-collection . #f)
      (-dof-fullgc-on-promotion . #f)
      (-np-remset-limit . #t)
      (-feeling-lucky . #t)
      (-phase-detection . #t)
      (-areas . #t)))

  (define keywords-to-flatten
    `((-rof-fullgc-on-collection . #f)
      (-rof-fullgc-on-promotion . #f)
      (-gen-fullgc-on-promotion . #f)))

  (define assocs '())

  (define (remove-from sys)
    (cond ((null? sys) '())
          ((assq (car sys) keywords-to-remove)
           =>
           (lambda (probe)
             (if (cdr probe)
                 (begin (set! assocs (cons (cons (car sys) (cadr sys))
                                           assocs))
                        (remove-from (cddr sys)))
                 (begin (set! assocs (cons (cons (car sys) #t)
                                           assocs))
                        (remove-from (cdr sys))))))
          ((assq (car sys) keywords-to-flatten)
           =>
           (lambda (probe)
             (if (cdr probe)
                 (begin (set! assocs (cons (cons (car sys) (cadr sys))
                                           assocs))
                        (cons (car sys)
                              (remove-from (cddr sys))))
                 (begin (set! assocs (cons (cons (car sys) #t)
                                           assocs))
                        (cons (car sys)
                              (remove-from (cdr sys)))))))
          (else
           (cons (car sys)
                 (remove-from (cdr sys))))))

  (define (less? a b)
    (list<? (cons (cdr (assq 'benchmark-name a))
                  (cdr (assq 'system-info a)))
            (cons (cdr (assq 'benchmark-name b))
                  (cdr (assq 'system-info b)))))


  (sort (map (lambda (system)
               (let ((system-info (assq 'system-info system)))
                 (set! assocs '())
                 (set-cdr! system-info (remove-from (cdr system-info)))
                 (for-each 
                  (lambda (key)
                    (if (and (not (cdr key))
                             (not (assq (car key) assocs)))
                        (set! assocs (cons (cons (car key) #f)
                                           assocs))))
                  keywords-to-remove)
                 (append system assocs)))
             data)
        less?))


; Data are as produced by basic-statistics.
;
; Compares groups of runs that have comparable headings (see below)
; and retains the most desirable.  The resulting list of data
; are sorted by heading.
;
; more-desirable takes two data from the input and returns #t if the
; first is more desirable (by whatever criterion) than the second,
; and #f otherwise.
;
; A heading is a list that consists of benchmark name, collector, run
; type (fixed or variable), trigger (if fixed), load (if variable), heap
; size (if fixed), and an "auxiliary" field, which consists of all options
; following the fixed options in the system-info field of the datum.  
; Note, the auxiliary field will introduce a lot of noise if the data
; have not been cleaned by remove-option-noise or flatten-option-noise.
;
; (It's called a heading because it has the same contents as the report 
; heading that's printed by STANDARD-REPORT.)

(define (select-most-desirable-run-and-resort data more-desirable?)

  (define current-system #f)
  (define output '())

  (for-each (lambda (x)
              (cond ((not current-system)
                     (set! current-system x))
                    ((equal? (standard-header current-system)
                             (standard-header x))
                     (if (more-desirable? x current-system)
                         (set! current-system x)))
                    (else
                     (set! output (cons current-system output))
                     (set! current-system x))))
            data)
  (set! output (cons current-system output))
  (sort-by-standard-header (reverse output)))


; Select-option returns a subset of the data that satisfy or do not
; satisfy a certain condition.
;
; Data is as produced by basic-statistics; the return value is
; a subsequence of this list.
;
; Sense is a boolean.  Return all entries for which (eqv? sense RESULT)
; when RESULT is as defined below.
;
; OPT is a symbol or a list; a symbol S is shorthand for the list (S).
;
; The entry for system-info in each datum is traversed left-to-right.  
; If a sequence of tokens that match the list OPT is found, then RESULT 
; is #T for the datum.  (EQV? is used as the equality test.)  Otherwise, 
; RESULT is #f for the datum.

(define (select-option opt sense data)
  (let ((opt (if (atom? opt) (list opt) opt))
        (meaning (if sense (lambda (x) x) not)))
    (filter (lambda (system)
              (let ((system-info (cdr (assq 'system-info system))))
                (let loop ((i system-info))
                  (cond ((null? i) (meaning #f))
                        ((eqv? (car i) (car opt))
                         (let loop2 ((j (cdr i)) (o (cdr opt)))
                           (cond ((null? o) (meaning #t))
                                 ((null? j) 
                                  (loop (cdr i)))
                                 ((eqv? (car j) (car o))
                                  (loop2 (cdr j) (cdr o)))
                                 (else
                                  (loop (cdr i))))))
                        (else
                         (loop (cdr i)))))))
            data)))


; data are as produced by basic-statistics.

(define (select-benchmark data name)
  (filter (lambda (datum)
            (equal? (benchmark-designator datum) name))
          data))

; data are as produced by basic-statistics.

(define (select-benchmarks data names)
  (mappend (lambda (n) (select-benchmark data n)) names))

; data are as produced by basic-statistics.

(define (select-if-present option+value data)
  (filter (lambda (x)
            (member option+value x))
          data))

; This tries to compare two lists in a rational way, if the lists have
; comparable data in matching locations.  Works on nested lists, but
; ignores the list structure: compares non-pairs only.

(define (list<? a b)

  (define (symbol<? a b)
    (string<? (symbol->string a) (symbol->string b)))

  (define (<? a b)
    (cond ((and (number? a) (number? b)) (< a b))
          ((and (symbol? a) (symbol? b)) (symbol<? a b))
          ((and (boolean? a) (boolean? b)) (and (not a) b))
          (else
           (string<? (cond ((string? a) a)
                           ((symbol? a) (symbol->string a))
                           ((number? a) (number->string a))
                           (else (error "<? " a)))
                     (cond ((string? b) b)
                           ((symbol? b) (symbol->string b))
                           ((number? b) (number->string b))
                           (else (error "<? " b)))))))

  (define (=? a b)
    (cond ((and (number? a) (number? b)) (= a b))
          ((and (symbol? a) (symbol? b)) (eq? a b))
          ((and (boolean? a) (boolean? b)) (eq? a b))
          (else
           (string=? (cond ((string? a) a)
                           ((symbol? a) (symbol->string a))
                           ((number? a) (number->string a))
                           (else (error "=? " a)))
                     (cond ((string? b) b)
                           ((symbol? b) (symbol->string b))
                           ((number? b) (number->string b))
                           (else (error "=? " b)))))))

  (define (less? a b)
    (cond ((null? a) (not (null? b)))
          ((null? b) #f)
          ((<? (car a) (car b)))
          ((=? (car a) (car b))
           (less? (cdr a) (cdr b)))
          (else #f)))

  (less? (flatten a) (flatten b)))


(define (substitute-char old new s)
  (list->string (map (lambda (x) (if (char=? x old) new x))
                     (string->list s))))


; eof
